{$INCLUDE cHeader.inc}
unit cDateTime;

interface



{                                                                              }
{ DateTime unit v0.02 (L0)                                                     }
{                                                                              }
{   A collection of date/time functions.                                       }
{                                                                              }
{                                                                              }
{ This unit is copyrighted  1999 by David Butler (david@e.co.za)              }
{                                                                              }
{ I invite you to use this unit, free of charge.                               }
{ I invite you to distibute this unit, but it must be for free.                }
{ I also invite you to contribute to its development, but do not distribute    }
{ a modified copy of this file, send modifications, suggestions and bug        }
{ reports to david@e.co.za                                                     }
{                                                                              }
{                                                                              }
{ A good source of information on calendars is the FAQ ABOUT CALENDARS,        }
{ available at http://www.tondering.dk/claus/calendar.html                     }
{                                                                              }
{ Note the following (and more) is available in SysUtils:                      }
{   Function IsLeapYear (Year : Word) : Boolean                                }
{     (1 = Sunday .. 7 = Saturday)                                             }
{   Function EncodeDate(Year, Month, Day: Word): TDateTime;                    }
{   Procedure DecodeDate (D : DateTime; var Year, Month, Day : Word);          }
{   var ShortDayNames, LongDayNames, ShortMonthNames, LongMonthNames : Array   }
{                                                                              }
{                                                                              }
{ Revision history:                                                            }
{   1999/11/10  v0.01  Initial version from scratch. Add functions. DayOfYear. }
{   1999/11/21  v0.02  EasterSunday function. Diff functions. ISOInteger.      }
{                                                                              }
{ To-do:                                                                       }
{   * ISOWeekNumber                                                            }
{                                                                              }


uses
     // System units
     SysUtils; // Exception, DateTime functions



type
     EDateTime = class(Exception);

     {                                                                              }
     { Decoding                                                                     }
     {                                                                              }
function Year(const D: TDateTime): Word;
function Month(const D: TDateTime): Word;
function Day(const D: TDateTime): Word;
function Hour(const D: TDateTime): Word;
function Minute(const D: TDateTime): Word;
function Second(const D: TDateTime): Word;
function Millisecond(const D: TDateTime): Word;

const
     OneDay = 1.0;
     OneHour = OneDay / 24.0;
     OneMinute = OneHour / 60.0;
     OneSecond = OneMinute / 60.0;
     OneMillisecond = OneSecond / 1000.0;



     {                                                                              }
     { Comparison                                                                   }
     {                                                                              }
function IsEqual(const D1, D2: TDateTime): Boolean; overload;
function IsEqual(const D1: TDateTime; const Ye, Mo, Da: Word): Boolean; overload;
function IsEqual(const D1: TDateTime; const Ho, Mi, Se, ms: Word): Boolean; overload;
function IsAM(const D: TDateTime): Boolean;
function IsPM(const D: TDateTime): Boolean;
function IsMidnight(const D: TDateTime): Boolean;
function IsNoon(const D: TDateTime): Boolean;
function IsSunday(const D: TDateTime): Boolean;
function IsMonday(const D: TDateTime): Boolean;
function IsTuesday(const D: TDateTime): Boolean;
function IsWedneday(const D: TDateTime): Boolean;
function IsThursday(const D: TDateTime): Boolean;
function IsFriday(const D: TDateTime): Boolean;
function IsSaturday(const D: TDateTime): Boolean;
function IsWeekend(const D: TDateTime): Boolean;



{                                                                              }
{ Relative date/times                                                          }
{                                                                              }
function Noon(const D: TDateTime): TDateTime;
function Midnight(const D: TDateTime): TDateTime;
function FirstDayOfMonth(const D: TDateTime): TDateTime;
function LastDayOfMonth(const D: TDateTime): TDateTime;
function NextWorkday(const D: TDateTime): TDateTime;
function PreviousWorkday(const D: TDateTime): TDateTime;
function FirstDayOfYear(const D: TDateTime): TDateTime;
function LastDayOfYear(const D: TDateTime): TDateTime;
function EasterSunday(const Year: Word): TDateTime;
function GoodFriday(const Year: Word): TDateTime;

function AddMilliseconds(const D: TDateTime; const N: Integer): TDateTime;
function AddSeconds(const D: TDateTime; const N: Integer): TDateTime;
function AddMinutes(const D: TDateTime; const N: Integer): TDateTime;
function AddHours(const D: TDateTime; const N: Integer): TDateTime;
function AddDays(const D: TDateTime; const N: Integer): TDateTime;
function AddWeeks(const D: TDateTime; const N: Integer): TDateTime;
function AddMonths(const D: TDateTime; const N: Integer): TDateTime;
function AddYears(const D: TDateTime; const N: Integer): TDateTime;



{                                                                              }
{ Counting                                                                     }
{   DayOfYear and WeekNumber start at 1.                                       }
{   WeekNumber is not the ISO week number but the week number where week one   }
{     starts at Jan 1.                                                         }
{   For reference: ISO standard 8601:1988 - (European Standard EN 28601).      }
{     "It states that a week is identified by its number in a given year.      }
{      A week begins with a Monday (day 1) and ends with a Sunday (day 7).     }
{      The first week of a year is the one which includes the first Thursday   }
{      (day 4), or equivalently the one which includes January 4.              }
{      In other words, the first week of a new year is the week that has the   }
{      majority of its days in the new year."                                  }
{                                                                              }
function DayOfYear(const Ye, Mo, Da: Word): Integer; overload;
function DayOfYear(const D: TDateTime): Integer; overload;
function DaysInMonth(const Ye, Mo: Word): Integer; overload;
function DaysInMonth(const D: TDateTime): Integer; overload;
function DaysInYear(const Ye: Word): Integer; overload;
function DaysInYear(const D: TDateTime): Integer; overload;
function WeekNumber(const D: TDateTime): Integer;



{                                                                              }
{ Difference                                                                   }
{                                                                              }
function DiffMilliseconds(const D1, D2: TDateTime): Integer;
function DiffSeconds(const D1, D2: TDateTime): Integer;
function DiffMinutes(const D1, D2: TDateTime): Integer;
function DiffHours(const D1, D2: TDateTime): Integer;
function DiffDays(const D1, D2: TDateTime): Integer;
function DiffWeeks(const D1, D2: TDateTime): Integer;
function DiffMonths(const D1, D2: TDateTime): Integer;
function DiffYears(const D1, D2: TDateTime): Integer;



{                                                                              }
{ Conversions                                                                  }
{   ANSI Integer is an integer in the format YYYYDDD (where DDD = day number)  }
{   ISO-8601 Integer date is an integer in the format YYYYMMDD.                }
{   RFC850 format "dddd, dd-mmm-yy hh:nn:ss GMT",                              }
{     eg "Monday, 02-Jun-82 12:01:03 GMT"                                      }
{   RFC1123 format "ddd, dd mmm yyyy hh:nn:ss GMT",                            }
{     eg "Mon, 02 Jun 1982 12:01:03 GMT"                                       }
{   TropicalYear is the time for one orbit of the earth around the sun.        }
{   SynodicMonth is the time between two full moons.                           }
{                                                                              }
function DateTimeToANSI(const D: TDateTime): Integer;
function ANSIToDateTime(const Julian: Integer): TDateTime;
function DateTimeToISOInteger(const D: TDateTime): Integer;
function DateTimeToISO(const D: TDateTime): string;
function ISOIntegerToDateTime(const ISOInteger: Integer): TDateTime;
function TimeToRFCTime(const D: TDateTime): string;
function DateTimeToRFC850(const D: TDateTime): string;
function DateTimeToRFC1123(const D: TDateTime): string;

const
     TropicalYear = 365.24219 * OneDay;
     SynodicMonth = 29.53059 * OneDay;


implementation



uses
     // System units
     Math,
     // L0 units
     cStrings;


{                                                                              }
{ Decoding                                                                     }
{                                                                              }

function Year(const D: TDateTime): Word;
var Mo, Da: Word;
begin
     DecodeDate(D, Result, Mo, Da);
end;

function Month(const D: TDateTime): Word;
var Ye, Da: Word;
begin
     DecodeDate(D, Ye, Result, Da);
end;

function Day(const D: TDateTime): Word;
var Ye, Mo: Word;
begin
     DecodeDate(D, Ye, Mo, Result);
end;

function Hour(const D: TDateTime): Word;
var Mi, Se, MS: Word;
begin
     DecodeTime(D, Result, Mi, Se, MS);
end;

function Minute(const D: TDateTime): Word;
var Ho, Se, MS: Word;
begin
     DecodeTime(D, Ho, Result, Se, MS);
end;

function Second(const D: TDateTime): Word;
var Ho, Mi, MS: Word;
begin
     DecodeTime(D, Ho, Mi, Result, MS);
end;

function Millisecond(const D: TDateTime): Word;
var Ho, Mi, Se: Word;
begin
     DecodeTime(D, Ho, Mi, Se, Result);
end;



{                                                                              }
{ Comparison                                                                   }
{                                                                              }

function IsEqual(const D1, D2: TDateTime): Boolean;
begin
     Result := Abs(D1 - D2) < OneMillisecond;
end;

function IsEqual(const D1: TDateTime; const Ye, Mo, Da: Word): Boolean;
var Ye1, Mo1, Da1: Word;
begin
     DecodeDate(D1, Ye1, Mo1, Da1);
     Result := (Da = Da1) and (Mo = Mo1) and (Ye = Ye1);
end;

function IsEqual(const D1: TDateTime; const Ho, Mi, Se, ms: Word): Boolean;
var Ho1, Mi1, Se1, ms1: Word;
begin
     DecodeTime(D1, Ho1, Mi1, Se1, ms1);
     Result := (ms = ms1) and (Se = Se1) and (Mi = Mi1) and (Ho = Ho1);
end;

function IsAM(const D: TDateTime): Boolean;
begin
     Result := Frac(D) < 0.5;
end;

function IsPM(const D: TDateTime): Boolean;
begin
     Result := Frac(D) >= 0.5;
end;

function IsNoon(const D: TDateTime): Boolean;
begin
     Result := Abs(Frac(D) - 0.5) < OneMillisecond;
end;

function IsMidnight(const D: TDateTime): Boolean;
var T: TDateTime;
begin
     T := Frac(D);
     Result := (T < OneMillisecond) or (T > 1.0 - OneMillisecond);
end;

function IsSunday(const D: TDateTime): Boolean;
begin
     Result := DayOfWeek(D) = 1;
end;

function IsMonday(const D: TDateTime): Boolean;
begin
     Result := DayOfWeek(D) = 2;
end;

function IsTuesday(const D: TDateTime): Boolean;
begin
     Result := DayOfWeek(D) = 3;
end;

function IsWedneday(const D: TDateTime): Boolean;
begin
     Result := DayOfWeek(D) = 4;
end;

function IsThursday(const D: TDateTime): Boolean;
begin
     Result := DayOfWeek(D) = 5;
end;

function IsFriday(const D: TDateTime): Boolean;
begin
     Result := DayOfWeek(D) = 6;
end;

function IsSaturday(const D: TDateTime): Boolean;
begin
     Result := DayOfWeek(D) = 7;
end;

function IsWeekend(const D: TDateTime): Boolean;
begin
     Result := DayOfWeek(D) in [1, 7];
end;

function IsWeekday(const D: TDateTime): Boolean;
begin
     Result := DayOfWeek(D) in [2..6];
end;



{                                                                              }
{ Relative calculations                                                        }
{                                                                              }

function Noon(const D: TDateTime): TDateTime;
begin
     Result := Int(D) + 0.5 * OneDay;
end;

function Midnight(const D: TDateTime): TDateTime;
begin
     Result := Int(D);
end;

function NextWorkday(const D: TDateTime): TDateTime;
begin
     case DayOfWeek(D) of
          1..5: Result := Trunc(D) + OneDay; // Sun..Thu
          6: Result := Trunc(D) + 3 * OneDay; // Fri
          7: Result := Trunc(D) + 2 * OneDay; // Sat
     else result := 0;
     end;
end;

function PreviousWorkday(const D: TDateTime): TDateTime;
begin
     case DayOfWeek(D) of
          1: Result := Trunc(D) - 2 * OneDay; // Sun
          2: Result := Trunc(D) - 3 * OneDay; // Mon
          3..7: Result := Trunc(D) - OneDay; // Tue-Sat
     else result := 0;
     end;
end;

function LastDayOfMonth(const D: TDateTime): TDateTime;
var Ye, Mo, Da: Word;
begin
     DecodeDate(D, Ye, Mo, Da);
     Result := EncodeDate(Ye, Mo, DaysInMonth(Ye, Mo));
end;

function FirstDayOfMonth(const D: TDateTime): TDateTime;
var Ye, Mo, Da: Word;
begin
     DecodeDate(D, Ye, Mo, Da);
     Result := EncodeDate(Ye, Mo, 1);
end;

function LastDayOfYear(const D: TDateTime): TDateTime;
var Ye, Mo, Da: Word;
begin
     DecodeDate(D, Ye, Mo, Da);
     Result := EncodeDate(Ye, 12, 31);
end;

function FirstDayOfYear(const D: TDateTime): TDateTime;
var Ye, Mo, Da: Word;
begin
     DecodeDate(D, Ye, Mo, Da);
     Result := EncodeDate(Ye, 1, 1);
end;

{ This algorithm comes from http://www.tondering.dk/claus/calendar.html:       }
{ " This algorithm is based in part on the algorithm of Oudin (1940) as        }
{   quoted in "Explanatory Supplement to the Astronomical Almanac",            }
{   P. Kenneth Seidelmann, editor.                                             }
{   People who want to dig into the workings of this algorithm, may be         }
{   interested to know that                                                    }
{     G is the Golden Number-1                                                 }
{     H is 23-Epact (modulo 30)                                                }
{     I is the number of days from 21 March to the Paschal full moon           }
{     J is the weekday for the Paschal full moon (0=Sunday, 1=Monday,etc.)     }
{     L is the number of days from 21 March to the Sunday on or before         }
{       the Paschal full moon (a number between -6 and 28) "                   }

function EasterSunday(const Year: Word): TDateTime;
var C, I, J, H, G, L: Integer;
     D, M: Word;
begin
     G := Year mod 19;
     C := Year div 100;
     H := (C - C div 4 - (8 * C + 13) div 25 + 19 * G + 15) mod 30;
     I := H - (H div 28) * (1 - (H div 28) * (29 div (H + 1)) * ((21 - G) div 11));
     J := (Year + Year div 4 + I + 2 - C + C div 4) mod 7;
     L := I - J;
     M := 3 + (L + 40) div 44;
     D := L + 28 - 31 * (M div 4);
     Result := EncodeDate(Year, M, D);
end;

function GoodFriday(const Year: Word): TDateTime;
begin
     Result := EasterSunday(Year) - 2 * OneDay;
end;

function AddMilliseconds(const D: TDateTime; const N: Integer): TDateTime;
begin
     Result := D + OneMillisecond * N;
end;

function AddSeconds(const D: TDateTime; const N: Integer): TDateTime;
begin
     Result := D + OneSecond * N;
end;

function AddMinutes(const D: TDateTime; const N: Integer): TDateTime;
begin
     Result := D + OneMinute * N;
end;

function AddHours(const D: TDateTime; const N: Integer): TDateTime;
begin
     Result := D + OneHour * N;
end;

function AddDays(const D: TDateTime; const N: Integer): TDateTime;
begin
     Result := D + N;
end;

function AddWeeks(const D: TDateTime; const N: Integer): TDateTime;
begin
     Result := D + N * 7 * OneDay;
end;

function AddMonths(const D: TDateTime; const N: Integer): TDateTime;
var Ye, Mo, Da: Word;
begin
     DecodeDate(D, Ye, Mo, Da);
     Inc(Ye, N div 12);
     Inc(Mo, N mod 12);
     if Mo > 12 then
     begin
          Dec(Mo, 12);
          Inc(Ye);
     end else
          if Mo < 1 then
     begin
          Inc(Mo, 12);
          Dec(Ye);
     end;
     Da := Min(Da, DaysInMonth(Ye, Mo));
     Result := EncodeDate(Ye, Mo, Da) + Frac(D);
end;

function AddYears(const D: TDateTime; const N: Integer): TDateTime;
var Ye, Mo, Da: Word;
begin
     DecodeDate(D, Ye, Mo, Da);
     Inc(Ye, N);
     Da := Min(Da, DaysInMonth(Ye, Mo));
     Result := EncodeDate(Ye, Mo, Da);
end;




{                                                                              }
{ Counting                                                                     }
{                                                                              }
const
     DaysInNonLeapMonth: array[1..12] of Integer = (
          31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
     CumDaysInNonLeapMonth: array[1..12] of Integer = (
          0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334);

function DayOfYear(const Ye, Mo, Da: Word): Integer; overload;
begin
     Result := CumDaysInNonLeapMonth[Mo] + Da;
     if (Mo > 2) and IsLeapYear(Ye) then
          Inc(Result);
end;

function DayOfYear(const D: TDateTime): Integer; overload;
var Ye, Mo, Da: Word;
begin
     DecodeDate(D, Ye, Mo, Da);
     Result := DayOfYear(Ye, Mo, Da);
end;

function DaysInMonth(const Ye, Mo: Word): Integer;
begin
     Result := DaysInNonLeapMonth[Mo];
     if (Mo = 2) and IsLeapYear(Ye) then
          Inc(Result);
end;

function DaysInMonth(const D: TDateTime): Integer;
var Ye, Mo, Da: Word;
begin
     DecodeDate(D, Ye, Mo, Da);
     Result := DaysInMonth(Ye, Mo);
end;

function DaysInYear(const Ye: Word): Integer;
begin
     if IsLeapYear(Ye) then
          Result := 366 else
          Result := 365;
end;

function DaysInYear(const D: TDateTime): Integer;
var Ye, Mo, Da: Word;
begin
     DecodeDate(D, Ye, Mo, Da);
     Result := DaysInYear(Ye);
end;

function WeekNumber(const D: TDateTime): Integer;
begin
     Result := (DiffDays(FirstDayOfYear(D), D) div 7) + 1;
end;



{                                                                              }
{ Difference                                                                   }
{                                                                              }

function DiffMilliseconds(const D1, D2: TDateTime): Integer;
begin
     Result := Trunc((D2 - D1) / OneMillisecond);
end;

function DiffSeconds(const D1, D2: TDateTime): Integer;
begin
     Result := Trunc((D2 - D1) / OneSecond);
end;

function DiffMinutes(const D1, D2: TDateTime): Integer;
begin
     Result := Trunc((D2 - D1) / OneMinute);
end;

function DiffHours(const D1, D2: TDateTime): Integer;
begin
     Result := Trunc((D2 - D1) / OneHour);
end;

function DiffDays(const D1, D2: TDateTime): Integer;
begin
     Result := Trunc(D2 - D1);
end;

function DiffWeeks(const D1, D2: TDateTime): Integer;
begin
     Result := Trunc(D2 - D1) div 7;
end;

function DiffMonths(const D1, D2: TDateTime): Integer;
var Ye1, Mo1, Da1: Word;
     Ye2, Mo2, Da2: Word;
     ModMonth1,
          ModMonth2: TDateTime;
begin
     DecodeDate(D1, Ye1, Mo1, Da1);
     DecodeDate(D2, Ye2, Mo2, Da2);
     Result := (Ye2 - Ye1) * 12 + (Mo2 - Mo1);
     ModMonth1 := Da1 + Frac(D1);
     ModMonth2 := Da2 + Frac(D2);
     if (D2 > D1) and (ModMonth2 < ModMonth1) then
          Dec(Result);
     if (D2 < D1) and (ModMonth2 > ModMonth1) then
          Inc(Result);
end;

function DiffYears(const D1, D2: TDateTime): Integer;
var Ye1, Mo1, Da1: Word;
     Ye2, Mo2, Da2: Word;
     ModYear1,
          ModYear2: TDateTime;
begin
     DecodeDate(D1, Ye1, Mo1, Da1);
     DecodeDate(D2, Ye2, Mo2, Da2);
     Result := Ye2 - Ye1;
     ModYear1 := Mo1 * 31 + Da1 + Frac(Da1);
     ModYear2 := Mo2 * 31 + Da2 + Frac(Da2);
     if (D2 > D1) and (ModYear2 < ModYear1) then
          Dec(Result);
     if (D2 < D1) and (ModYear2 > ModYear1) then
          Inc(Result);
end;




{                                                                              }
{ Conversions                                                                  }
{                                                                              }

function DateTimeToANSI(const D: TDateTime): Integer;
var Ye, Mo, Da: Word;
begin
     DecodeDate(D, Ye, Mo, Da);
     Result := Ye * 1000 + DayOfYear(Ye, Mo, Da);
end;

function ANSIToDateTime(const Julian: Integer): TDateTime;
var DDD, M, Y: Integer;
     I, C, J: Integer;
begin
     DDD := Julian mod 1000;
     if DDD = 0 then
          raise EDateTime.Create('Invalid ANSI date format');

     Y := Julian div 1000;
     M := 0;
     C := 0;
     for I := 1 to 12 do
     begin
          J := DaysInNonLeapMonth[I];
          if (I = 2) and IsLeapYear(Y) then
               Inc(J);
          Inc(C, J);
          if C >= DDD then
          begin
               M := I;
               break;
          end;
     end;
     if M = 0 then // DDD > end of year
          raise EDateTime.Create('Invalid ANSI date format');

     Result := EncodeDate(Y, M, DDD - C + J);
end;

function DateTimeToISOInteger(const D: TDateTime): Integer;
var Ye, Mo, Da: Word;
begin
     DecodeDate(D, Ye, Mo, Da);
     Result := Ye * 10000 + Mo * 100 + Da;
end;

function DateTimeToISO(const D: TDateTime): string;
var Ye, Mo, Da: Word;
begin
     DecodeDate(D, Ye, Mo, Da);
     Result := IntToStr(Ye) + '-' +
          PadLeft(IntToStr(Mo), '0', 2) + '-' +
          PadLeft(IntToStr(Da), '0', 2);
end;

function ISOIntegerToDateTime(const ISOInteger: Integer): TDateTime;
var Ye, Mo, Da: Word;
begin
     Ye := ISOInteger div 10000;
     Mo := (ISOInteger mod 10000) div 100;
     if (Mo < 1) or (Mo > 12) then
          raise EDateTime.Create('Invalid ISO Integer date format');
     Da := ISOInteger mod 100;
     if (Da < 1) or (Da > DaysInMonth(Ye, Mo)) then
          raise EDateTime.Create('Invalid ISO Integer date format');
     Result := EncodeDate(Ye, Mo, Da);
end;

function TimeToRFCTime(const D: TDateTime): string;
var Ho, Mi, Se, Ms: Word;
begin
     DecodeTime(D, Ho, Mi, Se, Ms);
     Result := PadLeft(IntToStr(Ho), '0', 2) + ':' +
          PadLeft(IntToStr(Mi), '0', 2) + ':' +
          PadLeft(IntToStr(Se), '0', 2) + ' GMT';
end;

const
     RFC850DayNames: array[1..7] of string = (
          'Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday');
     RFCMonthNames: array[1..12] of string = (
          'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec');

function DateTimeToRFC850(const D: TDateTime): string;
var Ye, Mo, Da: Word;
begin
     DecodeDate(D, Ye, Mo, Da);
     Result := RFC850DayNames[DayOfWeek(D)] + ', ' +
          PadLeft(IntToStr(Da), '0', 2) + '-' +
          RFCMonthNames[Mo] + '-' +
          PadLeft(IntToStr(Ye mod 100), '0', 2) + ' ' +
          TimeToRFCTime(D);
end;

const
     RFC1123DayNames: array[1..7] of string = (
          'Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat');

function DateTimeToRFC1123(const D: TDateTime): string;
var Ye, Mo, Da: Word;
begin
     DecodeDate(D, Ye, Mo, Da);
     Result := RFC1123DayNames[DayOfWeek(D)] + ', ' +
          PadLeft(IntToStr(Da), '0', 2) + ' ' +
          RFCMonthNames[Mo] + ' ' +
          IntToStr(Ye) + ' ' +
          TimeToRFCTime(D);
end;

end.

