unit cmplx;
//-------------------------------------------------------------------------//
//                 Complex numbers routines library                        //
//               Copyright (c) 2001 by Serghei Gaivan                      //
//                    e-mail: gaivan@mail.md                               //
//                  http://gaivan.hypermart.net                            //
//-------------------------------------------------------------------------//

interface

type
  complex = record
      x: real;
      y: real;
    end;

const
  ui: complex = (x: 0; y: -1);
  ur: complex = (x: 1; y: 0);

function Re(z: complex): real;          // z :--> Re(z)
function Im(z: complex): real;          // z :--> Im(z)

//------ Unary operations ----------------------------------------------------//
function Conj(z: complex): complex;     // z :--> z*
function Neg(z: complex): complex;      // z :--> -z
function Flip(z: complex): complex;     // (x,y) :--> (y,x)
function RCW(z: complex): complex;      // (x,y) :-->  (-y,x)
function RCC(z: complex): complex;      // (x,y) :-->  (y,-x)

//------ Binary operations ---------------------------------------------------//
function CSum(z1,z2: complex): complex; // z1,z2 :--> z1+z2
function CSub(z1,z2: complex): complex; // z1,z2 :--> z1-z2
function CMul(z1,z2: complex): complex; // z1,z2 :--> z1*z2
function CDiv(z1,z2: complex): complex; // z1,z2 :--> z1/z2

//------ Standard routines ---------------------------------------------------//
function Polar(rho,phi: real): complex; // (rho,phi) :--> z
function CAbs(z: complex): real;        // z :--> |z|
function Arg(z: complex): real;         // z :--> arg(z)
function Norm(z: complex): real;        // z :--> |z|^2

//------ Algebraic functions -------------------------------------------------//
function CSqr(z: complex): complex;     // z :--> z^2
function CInv(z: complex): complex;     // z :--> 1/z
function CSqrt(z: complex): complex;    // z :--> Sqrt(z)
function Pow(z: complex; n: integer): complex;  // z :--> z^n

//------ Transcendental functions --------------------------------------------//
function CLn(z: complex): complex;      // z :--> Ln(z)
function CExp(z: complex): complex;     // z :--> Exp(z)

//------ Trigonometric functions ---------------------------------------------//
function CSin(z: complex): complex;     // z :--> Sin(z)
function CCos(z: complex): complex;     // z :--> Cos(z)
function CTan(z: complex): complex;     // z :--> Tan(z)
function CCotan(z: complex): complex;   // z :--> Cotan(z)
function CAcos(z: complex): complex;    // z :--> Acos(z)
function CAsin(z: complex): complex;    // z :--> Asin(z)

//------ Hyperbolic functions ------------------------------------------------//
function CSinh(z: complex): complex;   // z :--> Sinh(z)
function CCosh(z: complex): complex;   // z :--> Cosh(z)
function CTanh(z: complex): complex;   // z :--> Tanh(z)
function CCotanh(z: complex): complex; // z :--> Cotanh(z)


implementation

{uses
    Math;
}

//----------------------------------------------------------------------------//

function Re(z: complex): real;
// z :--> Re(z)
asm
         FLD    z.x;
end;

//----------------------------------------------------------------------------//

function Im(z: complex): real;
// z :--> Im(z)
asm
         FLD    z.y;
end;

//----------------------------------------------------------------------------//
//------ Unary operations ----------------------------------------------------//
//----------------------------------------------------------------------------//

function Conj(z: complex): complex;
// z :--> z*
asm
         FLD     z.y
         FCHS
         FSTP    result.y
         FLD     z.x
         FSTP    result.x
end;

//----------------------------------------------------------------------------//

function Neg(z: complex): complex;
// (x,y) :--> (-x,-y)
asm
         FLD     z.x
         FCHS
         FSTP    result.x
         FLD     z.y
         FCHS
         FSTP    result.y
end;

//----------------------------------------------------------------------------//

function Flip(z: complex): complex;
// (x,y) :-->  (y,x)
asm
         FLD     z.y
         FSTP    result.x
         FLD     z.x
         FSTP    result.y
end;

//----------------------------------------------------------------------------//

function RCW(z: complex): complex;
// (x,y) :-->  (-y,x) that is z :--> i*z
asm
         FLD     z.y
         FCHS
         FSTP    result.x
         FLD     z.x
         FSTP    result.y
end;

//----------------------------------------------------------------------------//

function RCC(z: complex): complex;
// (x,y) :-->  (y,-x)
asm
         FLD     z.y
         FSTP    result.x
         FLD     z.x
         FCHS
         FSTP    result.y
end;

//----------------------------------------------------------------------------//
//------ Binary operations ---------------------------------------------------//
//----------------------------------------------------------------------------//

function CSum(z1,z2: complex): complex;
// z1,z2 :--> z1+z2
asm
         FLD      z1.x
         FLD      z2.x
         FADD
         FSTP     result.x
         FLD      z1.y
         FLD      z2.y
         FADD
         FSTP     result.y
end;

//----------------------------------------------------------------------------//

function CSub(z1,z2: complex): complex;
// z1,z2 :--> z1-z2
asm
         FLD      z1.x
         FLD      z2.x
         FSUB
         FSTP     result.x
         FLD      z1.y
         FLD      z2.y
         FSUB
         FSTP     result.y
end;

//----------------------------------------------------------------------------//

function CMul(z1,z2: complex): complex;
// z1,z2 :--> z1*z2
asm
         FLD      z1.x
         FLD      z2.x
         FLD      ST
         FMUL     ST,ST(2)
         FLD      z1.y
         FXCH     ST(1)
         FLD      z2.y
         FXCH     ST(1)
         FLD      ST(1)
         FMUL     ST,ST(3)
         FSUB
         FSTP     result.x
         FMULP    ST(3),ST(0)
         FMUL
         FADD
         FSTP     result.y
end;

//----------------------------------------------------------------------------//

function CDiv(z1,z2: complex): complex;
// z1,z2 :--> z1/z2
asm
         FLD      z2.y
         FLD      ST(0)
         FMUL     ST,ST
         FLD      z2.x
         FXCH     ST(1)
         FLD      ST(1)
         FMUL     ST,ST
         FADD
         FLD1
         FDIVR
         FLD      z1.x
         FLD      z1.y
         FXCH     ST(2)
         FLD      ST(1)
         FMUL     ST,ST(4)
         FLD      ST(3)        // y1 x1*x2 1/norm x1 y1 x2 y2
         FMUL     ST,ST(6)     // y1*y2 x1*x2 1/norm
         FADD
         FMUL     ST,ST(1)
         FSTP     result.x     // 1/norm x1 y1 x2 y2
         FXCH     ST(4)        // y2 x1 y1 x2 1/norm
         FMUL                  // x1*y2 y1 x2 1/norm
         FXCH     ST(2)        // x2 y1 x1*y2 1/norm
         FMUL                  // x2*y1 x1*y2 1/norm
         FSUBR
         FMUL
         FSTP     result.y
end;

//----------------------------------------------------------------------------//
//------ Standard routines ---------------------------------------------------//
//----------------------------------------------------------------------------//

function Polar(rho,phi: real): complex;
// (rho,phi) :--> z
asm
         FLD      rho
         FLD      phi
         FSINCOS
         FMUL     ST,ST(2)
         FSTP     result.x
         FMUL
         FSTP     result.y
end;

//----------------------------------------------------------------------------//

function CAbs(z: complex): real;
// z :--> |z|
asm
         FLD      z.y
         FMUL     ST,ST
         FLD      z.x
         FMUL     ST,ST
         FADD
         FSQRT
end;

//----------------------------------------------------------------------------//

function Arg(z: complex): real; 
// z :--> arg(z)
asm
         FLD     z.y
         FLD     z.x
         FPATAN
end;

//----------------------------------------------------------------------------//

function Norm(z: complex): real;
// z :--> |z|^2
asm
         FLD      z.y
         FMUL     ST,ST
         FLD      z.x
         FMUL     ST,ST
         FADD
end;

//----------------------------------------------------------------------------//
//------ Algebraic functions -------------------------------------------------//
//----------------------------------------------------------------------------//

function CSqr(z: complex): complex;
// z :--> z^2
asm
         FLD      z.y
         FLD      ST
         FMUL     ST,ST
         FLD      z.x
         FLD      ST
         FMUL     ST,ST
         FXCH     ST(3)
         FMUL
         FADD     ST,ST
         FSTP     result.y
         FSUB
         FSTP     result.x
end;

//----------------------------------------------------------------------------//

function CSqrt(z: complex): complex;
// z :--> sqrt(z)
asm
         FLD      z.x
         FLD      ST
         FMUL     ST,ST
         FLD      z.y
         FMUL     ST,ST
         FADD
         FSQRT
         FLD      ST(1)
         FADD     ST,ST(1)
         FABS
         FLD1
         FADD     ST,ST
         FDIV
         FSQRT
         FSTP     result.x
         FSUB
         FABS
         FLD1
         FADD     ST,ST
         FDIV
         FSQRT
         FSTP     result.y
end;

//----------------------------------------------------------------------------//

function CInv(z: complex): complex;
// z :--> 1/z
asm
         FLD      z.y
         FLD      ST
         FMUL     ST,ST
         FLD      z.x
         FXCH
         FLD      ST(1)
         FMUL     ST,ST
         FADD
         FLD1
         FDIVR
         FXCH     ST(2)
         FMUL     ST,ST(2)
         FSTP     result.y
         FMUL
         FSTP     result.x
end;

//----------------------------------------------------------------------------//

function Pow(z: complex; n: integer): complex;
// z :--> z^n
asm
         FLD      z.x
         FLD      z.y
         FLD1
         FLD      ST(2)
         FMUL     ST,ST
         FLD      ST(2)
         FMUL     ST,ST
         FADD
         FSQRT                        //sqrt(x*x+y*y) 1 y x
         MOV      EAX,n
         JMP      @2
  @1:    FMUL     ST,ST
  @2:    SHR      EAX,1
         JNC      @1
         FMUL     ST(1),ST
         JNZ      @1
         FSTP     ST(0)           // |z|^n y x
         FXCH     ST(2)           // x y |z|^n
         FPATAN                   // phi |z|^n
         MOV      [ESP-$04],n
         FILD     dword ptr [ESP-$04]     // n phi |z|^n
         FMUL                             // n*phi |z|^n
         FSINCOS                          // cos sin |z|^n
         FMUL     ST,ST(2)
         FSTP     result.x
         FMUL
         FSTP     result.y
end;

//----------------------------------------------------------------------------//
//------- Transcendent functions ---------------------------------------------//
//----------------------------------------------------------------------------//

function CLn(z: complex): complex;
// z :--> Ln(z)
asm
         FLD      z.y
         FLD      z.x
         FLDLN2
         FLD1
         FADD     ST,ST
         FDIV
         FLD      ST(2)
         FMUL     ST,ST
         FLD      ST(2)
         FMUL     ST,ST
         FADD
         FYL2X
         FSTP     result.x
         FPATAN
         FSTP     result.y
end;

//----------------------------------------------------------------------------//

function CExp(z: complex): complex;
// z :--> Exp(z)
asm
         FLD       z.x
         FLDL2E
         FMUL
         FLD       ST(0)
         FRNDINT
         FSUB      ST(1), ST
         FXCH      ST(1)
         F2XM1
         FLD1
         FADD
         FSCALE
         FSTP      ST(1)
         FLD       z.y
         FSINCOS
         FMUL      ST,ST(2)
         FSTP      result.x
         FMUL
         FSTP      result.y
end;

//----------------------------------------------------------------------------//
//------ Trigonometric functions ---------------------------------------------//
//----------------------------------------------------------------------------//

function CSin(z: complex): complex;
// z :--> Sin(z)
asm
         FLD       z.y
         FLDL2E
         FMUL
         FLD       ST(0)
         FRNDINT
         FSUB      ST(1), ST
         FXCH      ST(1)
         F2XM1
         FLD1
         FADD
         FSCALE
         FSTP      ST(1)
         FLD1
         FLD       ST(1)
         FADD      ST,ST
         FDIV
         FXCH
         FLD1
         FADD      ST,ST
         FDIV
         FLD       z.x
         FSINCOS
         FLD       ST(2)
         FSUB      ST,ST(4)
         FMUL
         FSTP      result.y
         FXCH      ST(2)
         FADD
         FMUL
         FSTP      result.x
end;

//----------------------------------------------------------------------------//

function CCos(z: complex): complex;
// z :--> Cos(z)
asm
         FLD       z.y
         FLDL2E
         FMUL
         FLD       ST(0)
         FRNDINT
         FSUB      ST(1), ST
         FXCH      ST(1)
         F2XM1
         FLD1
         FADD
         FSCALE
         FSTP      ST(1)
         FLD1
         FLD       ST(1)
         FADD      ST,ST
         FDIV
         FXCH
         FLD1
         FADD      ST,ST
         FDIV
         FLD       z.x
         FSINCOS
         FLD       ST(2)
         FADD      ST,ST(4)
         FMUL
         FSTP      result.x
         FXCH      ST(2)
         FSUBR
         FMUL
         FSTP      result.y
end;

//----------------------------------------------------------------------------//

function CTan(z: complex): complex;
// z :--> Tan(z)
asm
         FLD       z.x
         FADD      ST,ST
         FLD       z.y
         FADD      ST,ST           // 2y 2x
         FLDL2E
         FMUL
         FLD       ST(0)
         FRNDINT
         FSUB      ST(1), ST
         FXCH      ST(1)
         F2XM1
         FLD1
         FADD
         FSCALE
         FSTP      ST(1)           // exp(2y) 2x
         FLD1                      // 1 exp(2y) 2x
         FDIV      ST(0),ST(1)     // exp(-2y) exp(2y) 2x
         FLD1
         FADD      ST,ST           // 2 exp(-2y) exp(2y) 2x
         FLD       ST(0)           // 2 2 exp(-2y) exp(2y) 2x
         FDIVP     ST(2),ST(0)     // 2 exp(-2y)/2 exp(2y) 2x
         FDIVP     ST(2),ST(0)     // exp(-2y)/2 exp(2y)/2 2x
         FLD       ST(1)           // exp(2y)/2 exp(-2y)/2 exp(2y)/2 2x
         FSUB      ST(0),ST(1)     // sinh(2y) exp(-2y)/2 exp(2y)/2 2x
         FXCH      ST(2)           // exp(2y)/2 exp(-2y)/2 sinh(2y) 2x
         FADD                      // cosh(2y) sinh(2y) 2x
         FXCH      ST(2)           // 2x sinh(2y) cosh(2y)
         FSINCOS                   // cos(2x) sin(2x) sinh(2y) cosh(2y)
         FADDP     ST(3),ST(0)     // sin(2x) sinh(2y) (cos+cosh)
         FDIV      ST(0),ST(2)
         FSTP      result.x        // sinh(2y) (cos+cosh)
         FDIVR
         FSTP      result.y
end;

//----------------------------------------------------------------------------//

function CCotan(z: complex): complex;
// z :--> Cotan(z)
asm
         FLD       z.x
         FADD      ST,ST
         FLD       z.y
         FADD      ST,ST           // 2y 2x
         FLDL2E
         FMUL
         FLD       ST(0)
         FRNDINT
         FSUB      ST(1), ST
         FXCH      ST(1)
         F2XM1
         FLD1
         FADD
         FSCALE
         FSTP      ST(1)           // exp(2y) 2x
         FLD1                      // 1 exp(2y) 2x
         FDIV      ST(0),ST(1)     // exp(-2y) exp(2y) 2x
         FLD1
         FADD      ST,ST           // 2 exp(-2y) exp(2y) 2x
         FLD       ST(0)           // 2 2 exp(-2y) exp(2y) 2x
         FDIVP     ST(2),ST(0)     // 2 exp(-2y)/2 exp(2y) 2x
         FDIVP     ST(2),ST(0)     // exp(-2y)/2 exp(2y)/2 2x
         FLD       ST(0)           // exp(-2y)/2 exp(-2y)/2 exp(2y)/2 2x
         FSUB      ST(0),ST(2)     // -sinh(2y) exp(-2y)/2 exp(2y)/2 2x
         FXCH      ST(2)           // exp(2y)/2 exp(-2y)/2 -sinh(2y) 2x
         FADD                      // cosh(2y) -sinh(2y) 2x
         FXCH      ST(2)           // 2x -sinh(2y) cosh(2y)
         FSINCOS                   // cos(2x) sin(2x) -sinh(2y) cosh(2y)
         FSUBP     ST(3),ST(0)     // sin(2x) -sinh(2y) (cos-cosh)
         FDIV      ST(0),ST(2)
         FSTP      result.x        // -sinh(2y) (cos-cosh)
         FDIVR
         FSTP      result.y
end;

//----------------------------------------------------------------------------//

function CAcos(z: complex): complex;
// z :--> Acos(z)
begin
  result := RCC(CLn(CSum(z,RCW(CSqrt(CSub(ur,CSqr(z)))))));
end;

//----------------------------------------------------------------------------//

function CAsin(z: complex): complex;
// z :--> Asin(z)
begin
  result := RCC(CLn(CSum(RCW(z),CSqrt(CSub(ur,CSqr(z))))));
end;

//----------------------------------------------------------------------------//
//------ Hyperbolic functions -----------------------------------------------//
//----------------------------------------------------------------------------//

function CSinh(z: complex): complex;
// z :--> Sinh(z)
asm
         FLD       z.x
         FLDL2E
         FMUL
         FLD       ST(0)
         FRNDINT
         FSUB      ST(1), ST
         FXCH      ST(1)
         F2XM1
         FLD1
         FADD
         FSCALE
         FSTP      ST(1)
         FLD1
         FLD       ST(1)
         FADD      ST,ST
         FDIV
         FXCH
         FLD1
         FADD      ST,ST
         FDIV
         FLD       z.y
         FSINCOS
         FLD       ST(2)
         FSUB      ST,ST(4)
         FMUL
         FSTP      result.x
         FXCH      ST(2)
         FADD
         FMUL
         FSTP      result.y
end;

//----------------------------------------------------------------------------//

function CCosh(z: complex): complex;
// z :--> Cosh(z)
asm
         FLD       z.x
         FLDL2E
         FMUL
         FLD       ST(0)
         FRNDINT
         FSUB      ST(1), ST
         FXCH      ST(1)
         F2XM1
         FLD1
         FADD
         FSCALE
         FSTP      ST(1)
         FLD1
         FLD       ST(1)
         FADD      ST,ST
         FDIV
         FXCH
         FLD1
         FADD      ST,ST
         FDIV
         FLD       z.y
         FSINCOS
         FLD       ST(2)
         FADD      ST,ST(4)
         FMUL
         FSTP      result.x
         FXCH      ST(2)
         FSUB
         FMUL
         FSTP      result.y
end;

//----------------------------------------------------------------------------//

function CTanh(z: complex): complex;
// z :--> Tanh(z)
asm
         FLD       z.y
         FADD      ST,ST
         FLD       z.x
         FADD      ST,ST           // 2x 2y
         FLDL2E
         FMUL
         FLD       ST(0)
         FRNDINT
         FSUB      ST(1), ST
         FXCH      ST(1)
         F2XM1
         FLD1
         FADD
         FSCALE
         FSTP      ST(1)           // exp(2x) 2y
         FLD1                      // 1 exp(2x) 2y
         FDIV      ST(0),ST(1)     // exp(-2x) exp(2x) 2y
         FLD1
         FADD      ST,ST           // 2 exp(-2x) exp(2x) 2y
         FLD       ST(0)           // 2 2 exp(-2x) exp(2x) 2y
         FDIVP     ST(2),ST(0)     // 2 exp(-2x)/2 exp(2x) 2y
         FDIVP     ST(2),ST(0)     // exp(-2x)/2 exp(2x)/2 2y
         FLD       ST(1)           // exp(2x)/2 exp(-2x)/2 exp(2x)/2 2y
         FSUB      ST(0),ST(1)     // sinh(2x) exp(-2x)/2 exp(2x)/2 2y
         FXCH      ST(2)           // exp(2x)/2 exp(-2x)/2 sinh(2x) 2y
         FADD                      // cosh(2x) sinh(2x) 2y
         FXCH      ST(2)           // 2y sinh(2x) cosh(2x)
         FSINCOS                   // cos(2y) sin(2y) sinh(2x) cosh(2x)
         FADDP     ST(3),ST(0)     // sin(2y) sinh(2x) (cos+cosh)
         FDIV      ST(0),ST(2)
         FSTP      result.y        // sinh(2x) (cos+cosh)
         FDIVR
         FSTP      result.x
end;

//----------------------------------------------------------------------------//

function CCotanh(z: complex): complex;
// z :--> Cotanh(z)
asm
         FLD       z.y
         FADD      ST,ST
         FLD       z.x
         FADD      ST,ST
         FLDL2E
         FMUL
         FLD       ST(0)
         FRNDINT
         FSUB      ST(1), ST
         FXCH      ST(1)
         F2XM1
         FLD1
         FADD
         FSCALE
         FSTP      ST(1)
         FLD1
         FDIV      ST(0),ST(1)
         FLD1
         FADD      ST,ST
         FLD       ST(0)
         FDIVP     ST(2),ST(0)
         FDIVP     ST(2),ST(0)
         FLD       ST(0)
         FSUB      ST(0),ST(2)
         FXCH      ST(2)
         FADD
         FXCH      ST(2)
         FSINCOS
         FSUBRP    ST(3),ST(0)
         FDIV      ST(0),ST(2)
         FSTP      result.y
         FDIVR
         FSTP      result.x
end;


end. // end of module 
